@@ -9,6 +9,8 @@ header files (so that the JS compiler can see the constants in those
 headers, for the libc implementation in JS).
 """
 
+import glob
+import hashlib
 import json
 import logging
 import os
@@ -20,9 +22,11 @@ import textwrap
 
 from tools import (
   building,
+  cache,
   config,
   diagnostics,
   extract_metadata,
+  filelock,
   js_manipulation,
   shared,
   utils,
@@ -31,7 +35,7 @@ from tools import (
 from tools.settings import settings, user_settings
 from tools.shared import DEBUG, asmjs_mangle, in_temp
 from tools.toolchain_profiler import ToolchainProfiler
-from tools.utils import exit_with_error, path_from_root, removeprefix
+from tools.utils import exit_with_error, path_from_root
 
 sys.path.append(path_from_root('third_party'))
 import leb128
@@ -112,18 +116,21 @@ def update_settings_glue(wasm_file, metadata, base_metadata):
     settings.WASM_EXPORTS = base_metadata.all_exports
   else:
     settings.WASM_EXPORTS = metadata.all_exports
-  settings.DATA_EXPORTS = list(metadata.data_exports.keys())
   settings.HAVE_EM_ASM = bool(settings.MAIN_MODULE or len(metadata.em_asm_consts) != 0)
 
+  if settings.MAIN_MODULE and settings.ASYNCIFY == 1:
+    # These will be exported from Wasm, but only once we run the asyncify pass.
+    settings.WASM_EXPORTS += ['__asyncify_state', '__asyncify_data']
+
   # start with the MVP features, and add any detected features.
-  settings.BINARYEN_FEATURES = ['--mvp-features'] + metadata.features
+  building.binaryen_features = ['--mvp-features'] + metadata.features
   if settings.ASYNCIFY == 2:
-    settings.BINARYEN_FEATURES += ['--enable-reference-types']
+    building.binaryen_features += ['--enable-reference-types']
 
   if settings.PTHREADS:
-    assert '--enable-threads' in settings.BINARYEN_FEATURES
+    assert '--enable-threads' in building.binaryen_features
   if settings.MEMORY64:
-    assert '--enable-memory64' in settings.BINARYEN_FEATURES
+    assert '--enable-memory64' in building.binaryen_features
 
   settings.HAS_MAIN = bool(settings.MAIN_MODULE) or settings.PROXY_TO_PTHREAD or settings.STANDALONE_WASM or 'main' in settings.WASM_EXPORTS or '__main_argc_argv' in settings.WASM_EXPORTS
   if settings.HAS_MAIN and not settings.MINIMAL_RUNTIME:
@@ -170,6 +177,34 @@ def apply_static_code_hooks(forwarded_json, code):
   return code
 
 
+@ToolchainProfiler.profile()
+def generate_js_compiler_input_hash(symbols_only=False):
+  # We define a cache hit as when all the settings and all the `--js-library`
+  # contents are identical.
+  # Ignore certain settings that are not relevant to library deps.  Here we
+  # skip PRE_JS_FILES/POST_JS_FILES which don't affect the library symbol list
+  # and can contain full paths to temporary files.
+  skip_settings = {'PRE_JS_FILES', 'POST_JS_FILES'}
+  file_contents = [json.dumps(settings.external_dict(skip_keys=skip_settings), sort_keys=True, indent=2)]
+
+  files = glob.glob(utils.path_from_root('src/lib') + '/lib*.js')
+  # Also, include the js compiler code itself, in case it gets locally modified.
+  files += glob.glob(utils.path_from_root('src/*.mjs'))
+  files += settings.JS_LIBRARIES
+  if not symbols_only:
+    files += settings.PRE_JS_FILES
+    files += settings.POST_JS_FILES
+    # Also include other .js files that could be included in the output.
+    files += glob.glob(utils.path_from_root('src/*.js'))
+
+  for file in sorted(files):
+    file_contents.append(utils.read_file(file))
+
+  content = '\n'.join(file_contents)
+  content_hash = hashlib.sha1(content.encode('utf-8')).hexdigest()
+  return content_hash
+
+
 @ToolchainProfiler.profile()
 def compile_javascript(symbols_only=False):
   stderr_file = os.environ.get('EMCC_STDERR_FILE')
@@ -186,15 +221,8 @@ def compile_javascript(symbols_only=False):
   args = ['-']
   if symbols_only:
     args += ['--symbols-only']
-  out = shared.run_js_tool(path_from_root('tools/compiler.mjs'),
-                           args, input=settings_json, stdout=subprocess.PIPE, stderr=stderr_file)
-  if symbols_only:
-    glue = None
-    forwarded_data = out
-  else:
-    assert '//FORWARDED_DATA:' in out, 'Did not receive forwarded data in pre output - process failed?'
-    glue, forwarded_data = out.split('//FORWARDED_DATA:')
-  return glue, forwarded_data
+  return shared.run_js_tool(path_from_root('tools/compiler.mjs'),
+                            args, input=settings_json, stdout=subprocess.PIPE, stderr=stderr_file)
 
 
 def set_memory(static_bump):
@@ -234,10 +262,12 @@ def report_missing_exports(js_symbols):
     # maximum compatibility.
     return
 
-  if settings.EXPECT_MAIN and 'main' not in settings.WASM_EXPORTS and '__main_argc_argv' not in settings.WASM_EXPORTS:
-    # For compatibility with the output of wasm-ld we use the same wording here in our
-    # error message as if wasm-ld had failed.
-    exit_with_error('entry symbol not defined (pass --no-entry to suppress): main')
+  if settings.EXPECT_MAIN:
+    all_exports = settings.WASM_EXPORTS + settings.SIDE_MODULE_EXPORTS
+    if 'main' not in all_exports and '__main_argc_argv' not in all_exports:
+      # For compatibility with the output of wasm-ld we use the same wording here in our
+      # error message as if wasm-ld had failed.
+      exit_with_error('entry symbol not defined (pass --no-entry to suppress): main')
 
 
 # Test if the parentheses at body[openIdx] and body[closeIdx] are a match to
@@ -270,28 +300,59 @@ def trim_asm_const_body(body):
   return body
 
 
-def create_data_exports(data_exports):
-  lines = []
-  for k, v in data_exports.items():
-    if shared.is_internal_global(k):
-      continue
+def get_cached_file(filetype, filename, generator, cache_limit):
+  """This function implements a file cache which lives inside the main
+  emscripten cache directory but uses a per-file lock rather than a
+  cache-wide lock.
 
-    v = int(v)
+  The cache is pruned (by removing the oldest files) if it grows above
+  a certain number of files.
+  """
+  root = cache.get_path(filetype)
+  utils.safe_ensure_dirs(root)
 
-    if not settings.MEMORY64:
-      # We assume that global exports are addresses, which need to be interpreted as unsigned.
-      # This is not necessary (and does not work) under wasm64 when the globals are i64.
-      v = v & 0xFFFFFFFF
+  cache_file = os.path.join(root, filename)
 
-    if settings.RELOCATABLE:
-      v += settings.GLOBAL_BASE
-    mangled = asmjs_mangle(k)
-    if should_export(mangled) and not settings.MINIMAL_RUNTIME:
-      lines.append("var %s = Module['%s'] = %s;" % (mangled, mangled, v))
+  with filelock.FileLock(cache_file + '.lock'):
+    if os.path.exists(cache_file):
+      # Cache hit, read the file
+      file_content = utils.read_file(cache_file)
     else:
-      lines.append("var %s = %s;" % (mangled, v))
+      # Cache miss, generate the symbol list and write the file
+      file_content = generator()
+      utils.write_file(cache_file, file_content)
 
-  return '\n'.join(lines)
+  if len([f for f in os.listdir(root) if not f.endswith('.lock')]) > cache_limit:
+    with filelock.FileLock(cache.get_path(f'{filetype}.lock')):
+      files = []
+      for f in os.listdir(root):
+        if not f.endswith('.lock'):
+          f = os.path.join(root, f)
+          files.append((f, os.path.getmtime(f)))
+      files.sort(key=lambda x: x[1])
+      # Delete all but the newest N files
+      for f, _ in files[:-cache_limit]:
+        with filelock.FileLock(f + '.lock'):
+          utils.delete_file(f)
+
+  return file_content
+
+
+@ToolchainProfiler.profile()
+def compile_javascript_cached():
+  # Avoiding using the cache when generating struct info since
+  # this step is performed while the cache is locked.
+  # Sadly we have to skip the caching whenever we have user JS libraries.  This is because
+  # these libraries can import arbitrary other JS files (either vis node's `import` or via #include)
+  if DEBUG or settings.BOOTSTRAPPING_STRUCT_INFO or config.FROZEN_CACHE or settings.JS_LIBRARIES:
+    return compile_javascript()
+
+  content_hash = generate_js_compiler_input_hash()
+
+  # Limit of the overall size of the cache.
+  # This code will get test coverage since a full test run of `other` or `core`
+  # generates ~1000 unique outputs.
+  return get_cached_file('js_output', f'{content_hash}.js', compile_javascript, cache_limit=500)
 
 
 def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadata=None):
@@ -334,7 +395,7 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
   em_js_funcs = create_em_js(metadata)
 
   if settings.SIDE_MODULE:
-    # When building side modules, valid the EM_ASM and EM_JS string by running
+    # When building side modules, validate the EM_ASM and EM_JS string by running
     # them through node.  Without this step, syntax errors are not surfaced
     # until runtime.
     # We use subprocess directly here rather than shared.check_call since
@@ -358,7 +419,7 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
 
   for e in settings.EXPORTED_FUNCTIONS:
     if not js_manipulation.isidentifier(e):
-      exit_with_error(f'invalid export name: "{e}"')
+      diagnostics.warning('js-compiler', f'export name is not a valid JS symbol: "{e}".  Use `Module` or `wasmExports` to access this symbol')
 
   # memory and global initializers
 
@@ -374,13 +435,12 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     if settings.INITIAL_TABLE == -1:
       settings.INITIAL_TABLE = dylink_sec.table_size + 1
 
-  if settings.MAIN_MODULE and settings.ASYNCIFY == 1:
-    metadata.imports += ['__asyncify_state', '__asyncify_data']
-
   if metadata.invoke_funcs:
     settings.DEFAULT_LIBRARY_FUNCS_TO_INCLUDE += ['$getWasmTableEntry']
 
-  glue, forwarded_data = compile_javascript()
+  out = compile_javascript_cached()
+  assert '//FORWARDED_DATA:' in out, 'Did not receive forwarded data in pre output - process failed?'
+  glue, forwarded_data = out.split('//FORWARDED_DATA:', 1)
 
   forwarded_json = json.loads(forwarded_data)
 
@@ -407,13 +467,11 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
   if base_metadata:
     function_exports = base_metadata.function_exports
     other_exports = base_metadata.other_exports
-    # We want the real values from the final metadata but we only want to
-    # include names from the base_metadata.  See phase_link() in link.py.
-    data_exports = {k: v for k, v in metadata.data_exports.items() if k in base_metadata.data_exports}
   else:
     function_exports = metadata.function_exports
     other_exports = metadata.other_exports
-    data_exports = metadata.data_exports
+
+  other_exports = [e for e in other_exports if not shared.is_internal_global(e[0].name)]
 
   if settings.ASYNCIFY == 1:
     function_exports['asyncify_start_unwind'] = webassembly.FuncType([webassembly.Type.I32], [])
@@ -422,7 +480,7 @@ def emscript(in_wasm, out_wasm, outfile_js, js_syms, finalize=True, base_metadat
     function_exports['asyncify_stop_rewind'] = webassembly.FuncType([], [])
 
   parts = [pre]
-  parts += create_module(metadata, function_exports, data_exports, other_exports,
+  parts += create_module(metadata, function_exports, other_exports,
                          forwarded_json['librarySymbols'], forwarded_json['nativeAliases'])
   parts.append(post)
   settings.ALIASES = list(forwarded_json['nativeAliases'].keys())
@@ -901,7 +959,7 @@ def install_debug_wrapper(sym):
   if sym.startswith(('__asan_', 'emscripten_stack_', '_emscripten_stack_')):
     return False
   # `__trap` can occur before the runtime is initialized since it is used in abort.
-  # `emscripten_get_sbrk_ptr` can be called prior to runtime initialzaion by
+  # `emscripten_get_sbrk_ptr` can be called prior to runtime initialization by
   # the dynamic linking code.
   return sym not in ['__trap', 'emscripten_get_sbrk_ptr']
 
@@ -915,7 +973,7 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   receiving = ['\n// Imports from the Wasm binary.']
 
   if settings.WASM_ESM_INTEGRATION:
-    exported_symbols = other_exports + list(function_exports.keys())
+    exported_symbols = [e[0].name for e in other_exports] + list(function_exports.keys())
     exports = []
     for sym in exported_symbols:
       mangled = asmjs_mangle(sym)
@@ -935,7 +993,7 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
       receiving.append('  // Construct dynCalls mapping')
       for sym in function_exports:
         if sym.startswith('dynCall_'):
-          sig_str = sym.replace('dynCall_', '')
+          sig_str = sym.removeprefix('dynCall_')
           receiving.append(f"  dynCalls['{sig_str}'] = {sym};")
       receiving.append('}')
 
@@ -952,8 +1010,8 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   # function assignWasmExports(wasmExport) {
   #   _main = wasmExports["_main"];
   exports = {name: sig for name, sig in function_exports.items() if name != building.WASM_CALL_CTORS}
-  for name in other_exports:
-    exports[name] = None
+  for export, info in other_exports:
+    exports[export.name] = (export, info)
 
   mangled = [asmjs_mangle(s) for s in exports] + list(aliases.keys())
   if settings.ASSERTIONS:
@@ -961,10 +1019,16 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
     # folks try to call/use a reference that was taken before the
     # wasm module is available.
     for sym in mangled:
-      assignment = sym
-      if (settings.MODULARIZE or not settings.MINIMAL_RUNTIME) and should_export(sym) and settings.MODULARIZE != 'instance':
-        assignment += f" = Module['{sym}']"
-      receiving.append(f"var {assignment} = makeInvalidEarlyAccess('{sym}');")
+      module_export = (settings.MODULARIZE or not settings.MINIMAL_RUNTIME) and should_export(sym) and settings.MODULARIZE != 'instance'
+      if not js_manipulation.isidentifier(sym) and not module_export:
+        continue
+      assignment = f'var {sym}'
+      if module_export:
+        if js_manipulation.isidentifier(sym):
+          assignment += f" = Module['{sym}']"
+        else:
+          assignment = f"Module['{sym}']"
+      receiving.append(f"{assignment} = makeInvalidEarlyAccess('{sym}');")
   else:
     # Declare all exports in a single var statement
     sep = ',\n  '
@@ -984,25 +1048,48 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
 
   do_module_exports = (settings.MODULARIZE or not settings.MINIMAL_RUNTIME) and settings.MODULARIZE != 'instance'
   receiving.append('\nfunction assignWasmExports(wasmExports) {')
-  for sym, sig in exports.items():
-    is_function = sig is not None
+  if settings.ASSERTIONS:
+    for sym in exports:
+      if settings.EMBIND_GEN_MODE and sym.startswith('asyncify_'):
+        # EMBIND_GEN_MODE is run before binaryen so the asyncify exports that
+        # are created by binaryen will be missing.
+        continue
+      receiving.append(f"  assert(typeof wasmExports['{sym}'] != 'undefined', 'missing Wasm export: {sym}');")
+  for sym, info in exports.items():
+    is_function = type(info) == webassembly.FuncType
     mangled = asmjs_mangle(sym)
     assignment = mangled
     if generate_dyncall_assignment and is_function and sym.startswith('dynCall_'):
-      sig_str = sym.replace('dynCall_', '')
+      sig_str = sym.removeprefix('dynCall_')
       assignment += f" = dynCalls['{sig_str}']"
     if do_module_exports and should_export(mangled):
-      assignment += f" = Module['{mangled}']"
-    if settings.ASSERTIONS:
-      receiving.append(f"  assert(typeof wasmExports['{sym}'] != 'undefined', 'missing Wasm export: {sym}');")
+      if js_manipulation.isidentifier(mangled):
+        assignment += f" = Module['{mangled}']"
+      else:
+        assignment = f"Module['{mangled}']"
+    elif not js_manipulation.isidentifier(mangled):
+      # Symbol is not a valid JS identifier and also not exported. In this case we
+      # have nothing to do here.
+      continue
+
     if sym in alias_inverse_map:
       for target in alias_inverse_map[sym]:
         assignment += f" = {target}"
         if do_module_exports and target in settings.EXPORTED_RUNTIME_METHODS:
           assignment += f" = Module['{target}']"
     if is_function and install_debug_wrapper(sym):
-      nargs = len(sig.params)
+      nargs = len(info.params)
       receiving.append(f"  {assignment} = createExportWrapper('{sym}', {nargs});")
+    elif not is_function and info[0].kind == webassembly.ExternType.GLOBAL and not info[1].mutable:
+      if settings.LEGACY_VM_SUPPORT:
+        value = f"typeof wasmExports['{sym}'] == 'object' ? wasmExports['{sym}'].value : wasmExports['{sym}']"
+      else:
+        value = f"wasmExports['{sym}'].value"
+      if settings.MEMORY64:
+        value = f'Number({value})'
+      elif settings.CAN_ADDRESS_2GB:
+        value = f'({value}) >>> 0'
+      receiving.append(f"  {assignment} = {value};")
     else:
       receiving.append(f"  {assignment} = wasmExports['{sym}'];")
   receiving.append('}')
@@ -1010,10 +1097,9 @@ def create_receiving(function_exports, other_exports, library_symbols, aliases):
   return '\n'.join(receiving)
 
 
-def create_module(metadata, function_exports, data_exports, other_exports, library_symbols, aliases):
+def create_module(metadata, function_exports, other_exports, library_symbols, aliases):
   module = []
   module.append(create_receiving(function_exports, other_exports, library_symbols, aliases))
-  module.append(create_data_exports(data_exports))
 
   sending = create_sending(metadata, library_symbols)
   if settings.WASM_ESM_INTEGRATION:
@@ -1048,7 +1134,7 @@ def create_invoke_wrappers(metadata):
   """Asm.js-style exception handling: invoke wrapper generation."""
   invoke_wrappers = []
   for invoke in metadata.invoke_funcs:
-    sig = removeprefix(invoke, 'invoke_')
+    sig = invoke.removeprefix('invoke_')
     invoke_wrappers.append(js_manipulation.make_invoke(sig))
   return invoke_wrappers
 
@@ -1116,6 +1202,7 @@ def create_pointer_conversion_wrappers(metadata):
     '_emscripten_dlsync_self_async': '_p',
     '_emscripten_proxy_dlsync_async': '_pp',
     '_emscripten_wasm_worker_initialize': '_p_',
+    '_emscripten_proxy_poll_finish': '_pp_',
     '_wasmfs_rename': '_pp',
     '_wasmfs_readlink': '_pp',
     '_wasmfs_truncate': '_p_',
@@ -1145,7 +1232,7 @@ def create_pointer_conversion_wrappers(metadata):
     'emscripten_main_runtime_thread_id': 'p',
     '_emscripten_set_offscreencanvas_size_on_thread': '_pp__',
     'fileno': '_p',
-    '_emscripten_run_callback_on_thread': '_pp_pp',
+    '_emscripten_run_callback_on_thread': '_pp_ppp',
     '_emscripten_find_dylib': 'ppppp',
   }
 
@@ -1171,7 +1258,7 @@ function applySignatureConversions(wasmExports) {
     # dynCall_ functions are generated by binaryen.  They all take a
     # function pointer as their first argument.
     # The second part of this check can be removed once the table64
-    # llvm changs lands.
+    # llvm changes land.
     if symbol.startswith('dynCall_') and functype.params[0] == webassembly.Type.I64:
       sig = symbol.split('_')[-1]
       sig = ['p' if t == 'p' else '_' for t in sig]
